/******************************************************************************
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2023. All rights reserved.
 * iSulad licensed under the Mulan PSL v2.
 * You can use this software according to the terms and conditions of the Mulan PSL v2.
 * You may obtain a copy of Mulan PSL v2 at:
 *     http://license.coscl.org.cn/MulanPSL2
 * THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
 * PURPOSE.
 * See the Mulan PSL v2 for more details.
 * Author: wangfengtu
 * Create: 2020-10-19
 * Description: provide mount spec utils functions
 *******************************************************************************/

#define _GNU_SOURCE
#include "utils_mount_spec.h"

#include <stdlib.h>
#include <string.h>
#include <regex.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <strings.h>

#include "utils.h"
#include "utils_array.h"
#include "utils_string.h"
#include "path.h"

#define CACHE_ERRMSG_LEN 512
#define CACHE_ERRMSG(errmsg, fmt, args...)                     \
    do {                                                       \
        (void)snprintf(errmsg, CACHE_ERRMSG_LEN, fmt, ##args); \
    } while (0)

static int parse_mount_item_type(const char *value, char *mount_str, mount_spec *m, char *errmsg)
{
    if (m->type != NULL) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.More than one type found", mount_str);
        return EINVALIDARGS;
    }

    if (value == NULL || value[0] == 0) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Type is required", mount_str);
        return EINVALIDARGS;
    }

#ifdef ENABLE_OCI_IMAGE
    if (strcmp(value, MOUNT_TYPE_SQUASHFS) != 0 && strcmp(value, MOUNT_TYPE_BIND) != 0 &&
        strcmp(value, MOUNT_TYPE_VOLUME) != 0 && strcmp(value, MOUNT_TYPE_TMPFS) != 0) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Type must be one of squashfs/bind/volume/tmpfs",
                     mount_str);
#else
    if (strcmp(value, MOUNT_TYPE_SQUASHFS) && strcmp(value, MOUNT_TYPE_BIND)) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Type must be squashfs or bind", mount_str);
#endif
        return EINVALIDARGS;
    }

    m->type = util_strdup_s(value);

    return 0;
}

static int parse_mount_item_src(const char *value, char *mount_str, mount_spec *m, char *errmsg)
{
    char srcpath[PATH_MAX] = { 0 };

    /* If value of source is NULL, ignore it */
    if (value == NULL) {
        return 0;
    }

    if (m->source) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.More than one source found", mount_str);
        return EINVALIDARGS;
    }

#ifndef ENABLE_OCI_IMAGE
    if (value[0] != '/') {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Source must be absolute path", mount_str);
        return EINVALIDARGS;
    }
#endif

    if (value[0] == '/') {
        if (!util_clean_path(value, srcpath, sizeof(srcpath))) {
            CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Can't translate source path to clean path",
                         mount_str);
            return EINVALIDARGS;
        }
        m->source = util_strdup_s(srcpath);
    } else {
        m->source = util_strdup_s(value);
    }

    return 0;
}

static int parse_mount_item_dst(const char *value, char *mount_str, mount_spec *m, char *errmsg)
{
    char dstpath[PATH_MAX] = { 0 };

    /* If value of destination is NULL, ignore it */
    if (value == NULL) {
        return 0;
    }

    if (m->target) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.More than one destination found", mount_str);
        return EINVALIDARGS;
    }

    if (value[0] != '/') {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Destination must be absolute path", mount_str);
        return EINVALIDARGS;
    }

    if (!util_clean_path(value, dstpath, sizeof(dstpath))) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Can't translate destination path to clean path",
                     mount_str);
        return EINVALIDARGS;
    }

    if (strcmp(dstpath, "/") == 0) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Destination can't be '/'", mount_str);
        return EINVALIDARGS;
    }

    m->target = util_strdup_s(dstpath);

    return 0;
}

static int parse_mount_item_ro(const char *value, char *mount_str, mount_spec *m, char *errmsg)
{
    if (value == NULL || util_valid_value_true(value)) {
        m->readonly = true;
    } else if (util_valid_value_false(value)) {
        m->readonly = false;
    } else {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Invalid readonly mode:%s", mount_str, value);
        return EINVALIDARGS;
    }
    return 0;
}

static int parse_mount_item_propagation(const char *value, char *mount_str, mount_spec *m, char *errmsg)
{
    /* If value of destination is NULL, ignore it */
    if (value == NULL) {
        return 0;
    }

    if (m->bind_options != NULL && m->bind_options->propagation != NULL) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.More than one bind-propagation found", mount_str);
        return EINVALIDARGS;
    }

    if (!util_valid_propagation_mode(value)) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Invalid propagation mode:%s", mount_str, value);
        return EINVALIDARGS;
    }

    if (m->bind_options == NULL) {
        m->bind_options = util_common_calloc_s(sizeof(bind_options));
        if (m->bind_options == NULL) {
            CACHE_ERRMSG(errmsg, "Out of memory");
            return EINVALIDARGS;
        }
    }
    m->bind_options->propagation = util_strdup_s(value);

    return 0;
}

static int parse_mount_item_selinux(const char *value, char *mount_str, mount_spec *m, char *errmsg)
{
    /* If value of destination is NULL, ignore it */
    if (value == NULL) {
        return 0;
    }

    if (m->bind_options != NULL && m->bind_options->selinux_opts != NULL) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.More than one bind-selinux-opts found", mount_str);
        return EINVALIDARGS;
    }

    if (!util_valid_label_mode(value)) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Invalid bind selinux opts:%s", mount_str, value);
        return EINVALIDARGS;
    }

    if (m->bind_options == NULL) {
        m->bind_options = util_common_calloc_s(sizeof(bind_options));
        if (m->bind_options == NULL) {
            CACHE_ERRMSG(errmsg, "Out of memory");
            return EINVALIDARGS;
        }
    }
    m->bind_options->selinux_opts = util_strdup_s(value);

    return 0;
}

#ifdef ENABLE_OCI_IMAGE
static int parse_mount_item_tmpfs_size(const char *value, char *mount_str, mount_spec *m, char *errmsg)
{
    int64_t size = 0;

    /* If value of destination is NULL, ignore it */
    if (value == NULL) {
        return 0;
    }

    if (m->tmpfs_options != NULL && m->tmpfs_options->size_bytes != 0) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.More than one tmpfs-size found", mount_str);
        return EINVALIDARGS;
    }

    if (util_parse_byte_size_string(value, &size) != 0 || size < 0) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Invalid tmpfs-size:%s", mount_str, value);
        return EINVALIDARGS;
    }

    if (m->tmpfs_options == NULL) {
        m->tmpfs_options = util_common_calloc_s(sizeof(tmpfs_options));
        if (m->tmpfs_options == NULL) {
            CACHE_ERRMSG(errmsg, "Out of memory");
            return EINVALIDARGS;
        }
    }

    m->tmpfs_options->size_bytes = size;

    return 0;
}

static int parse_mount_item_tmpfs_mode(const char *value, char *mount_str, mount_spec *m, char *errmsg)
{
    uint32_t mode = 0;

    /* If value of destination is NULL, ignore it */
    if (value == NULL) {
        return 0;
    }

    if (m->tmpfs_options != NULL && m->tmpfs_options->mode != 0) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.More than one tmpfs-size found", mount_str);
        return EINVALIDARGS;
    }

    if (util_parse_octal_uint32(value, &mode) != 0) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Invalid tmpfs-mode:%s", mount_str, value);
        return EINVALIDARGS;
    }

    if (m->tmpfs_options == NULL) {
        m->tmpfs_options = util_common_calloc_s(sizeof(tmpfs_options));
        if (m->tmpfs_options == NULL) {
            CACHE_ERRMSG(errmsg, "Out of memory");
            return EINVALIDARGS;
        }
    }

    m->tmpfs_options->mode = mode;

    return 0;
}

static int parse_mount_item_nocopy(const char *value, char *mount_str, mount_spec *m, char *errmsg)
{
    /* If value of destination is NULL, ignore it */
    if (value == NULL) {
        return 0;
    }

    if (m->volume_options != NULL) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.More than one volume-nocopy found", mount_str);
        return EINVALIDARGS;
    }

    if (!util_valid_value_true(value) && !util_valid_value_false(value)) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Invalid volume nocopy:%s", mount_str, value);
        return EINVALIDARGS;
    }

    if (m->volume_options == NULL) {
        m->volume_options = util_common_calloc_s(sizeof(volume_options));
        if (m->volume_options == NULL) {
            CACHE_ERRMSG(errmsg, "Out of memory");
            return EINVALIDARGS;
        }
    }

    if (util_valid_value_true(value)) {
        m->volume_options->no_copy = true;
    } else {
        m->volume_options->no_copy = false;
    }

    return 0;
}
#endif

static bool exist_readonly_mode(char *mount_str)
{
    char tmp_mount_str[PATH_MAX] = { 0 };
    int sret = 0;

    // add "," at start and end of mount string to simplize check
    sret = snprintf(tmp_mount_str, sizeof(tmp_mount_str), ",%s,", mount_str);
    if (sret < 0 || (size_t)sret >= sizeof(tmp_mount_str)) {
        return false;
    }

    if (strstr(tmp_mount_str, ",ro,") != NULL || strstr(tmp_mount_str, ",readonly,") != NULL ||
        strstr(tmp_mount_str, ",ro=") != NULL || strstr(tmp_mount_str, ",readonly=") != NULL) {
        return true;
    }

    return false;
}

static bool valid_mount_spec_mode(char *mount_str, mount_spec *m, char *errmsg)
{
    if (strcmp(m->type, MOUNT_TYPE_VOLUME) == 0) {
        if (m->bind_options != NULL && m->bind_options->propagation != NULL) {
            CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Propagation must not be specified for type %s",
                         mount_str, m->type);
            return false;
        }
        if (exist_readonly_mode(mount_str) && m->source == NULL) {
            CACHE_ERRMSG(errmsg,
                         "Invalid mount specification '%s'.Readonly mode must not be specified for anonymous volume",
                         mount_str);
            return false;
        }
    }
    if (strcmp(m->type, MOUNT_TYPE_BIND) == 0 && m->volume_options != NULL) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.nocopy must not be specified for type %s", mount_str,
                     m->type);
        return false;
    }
    if (strcmp(m->type, MOUNT_TYPE_TMPFS) == 0) {
        if (m->volume_options != NULL) {
            CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Cannot mix volume options with type %s", mount_str,
                         m->type);
            return false;
        }
        if (m->bind_options != NULL) {
            CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Cannot mix bind options with type %s", mount_str,
                         m->type);
            return false;
        }
    }

    return true;
}

// check mount spec valid
static int check_mount_spec(char *mount_str, mount_spec *m, char *errmsg)
{
    // check source
    if (strcmp(m->type, MOUNT_TYPE_SQUASHFS) == 0 || strcmp(m->type, MOUNT_TYPE_BIND) == 0) {
        if (m->source == NULL) {
            CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Missing source", mount_str);
            return EINVALIDARGS;
        }
        if (m->source[0] != '/') {
            CACHE_ERRMSG(errmsg, "source %s should be absolute path for type %s", m->source, m->type);
            return -1;
        }
    }
    if (strcmp(m->type, MOUNT_TYPE_TMPFS) == 0 && m->source != NULL) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Source must not be specified", mount_str);
        return -1;
    }

    if (strcmp(m->type, MOUNT_TYPE_VOLUME) == 0 && m->source != NULL && !util_valid_volume_name(m->source)) {
        CACHE_ERRMSG(errmsg, "Invalid volume name %s, only \"%s\" are allowed", m->source, VALID_VOLUME_NAME);
        return -1;
    }

    // check destination
    if (m->target == NULL) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Missing destination", mount_str);
        return EINVALIDARGS;
    }

    if (m->target[0] != '/') {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.destination should be absolute path", mount_str);
        return -1;
    }

    if (strcmp(m->type, MOUNT_TYPE_SQUASHFS) == 0) {
        char real_path[PATH_MAX] = { 0 };
        if (strlen(m->source) > PATH_MAX || realpath(m->source, real_path) == NULL) {
            CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Source %s not exist", mount_str, m->source);
            return EINVALIDARGS;
        }

        /* Make sure it's a regular file */
        if (!util_valid_file(real_path, S_IFREG)) {
            CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Source %s is not a squashfs file", mount_str,
                         m->source);
            return EINVALIDARGS;
        }
    }

    if (!valid_mount_spec_mode(mount_str, m, errmsg)) {
        return -1;
    }

    return 0;
}

static int parse_mounts_item(const char *mntkey, const char *value, char *mount_str, mount_spec *m, char *errmsg)
{
    if (util_valid_key_type(mntkey)) {
        return parse_mount_item_type(value, mount_str, m, errmsg);
    } else if (util_valid_key_src(mntkey)) {
        return parse_mount_item_src(value, mount_str, m, errmsg);
    } else if (util_valid_key_dst(mntkey)) {
        return parse_mount_item_dst(value, mount_str, m, errmsg);
    } else if (util_valid_key_ro(mntkey)) {
        return parse_mount_item_ro(value, mount_str, m, errmsg);
    } else if (util_valid_key_propagation(mntkey)) {
        return parse_mount_item_propagation(value, mount_str, m, errmsg);
    } else if (util_valid_key_selinux(mntkey)) {
        return parse_mount_item_selinux(value, mount_str, m, errmsg);
#ifdef ENABLE_OCI_IMAGE
    } else if (util_valid_key_tmpfs_size(mntkey)) {
        return parse_mount_item_tmpfs_size(value, mount_str, m, errmsg);
    } else if (util_valid_key_tmpfs_mode(mntkey)) {
        return parse_mount_item_tmpfs_mode(value, mount_str, m, errmsg);
    } else if (util_valid_key_nocopy(mntkey)) {
        return parse_mount_item_nocopy(value, mount_str, m, errmsg);
#endif
    } else {
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Unsupported item:%s", mount_str, mntkey);
        return EINVALIDARGS;
    }
}

int util_parse_mount_spec(char *mount_str, mount_spec **spec, char **errmsg_out)
{
    mount_spec *m = NULL;
    int ret = 0;
    size_t i = 0;
    size_t items_len = 0;
    char **items = NULL;
    char **key_val = NULL;
    char errmsg[CACHE_ERRMSG_LEN] = { 0 };

    if (mount_str == NULL) {
        CACHE_ERRMSG(errmsg, "Invalid mount specification: can't be empty");
        return -1;
    }
    if (mount_str[0] == '\0') {
        CACHE_ERRMSG(errmsg, "Invalid mount specification: can't be empty");
        return -1;
    }

    if (spec == NULL) {
        CACHE_ERRMSG(errmsg, "Invalid spec: can't be NULL");
        return -1;
    }

    if (errmsg_out == NULL) {
        CACHE_ERRMSG(errmsg, "Invalid errmsg_out: can't be NULL");
        return -1;
    }

    m = util_common_calloc_s(sizeof(mount_spec));
    if (m == NULL) {
        CACHE_ERRMSG(errmsg, "out of memory");
        ret = -1;
        goto out;
    }

    items = util_string_split(mount_str, ',');
    if (items == NULL) {
        ret = EINVALIDARGS;
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'. unsupported format", mount_str);
        goto out;
    }

    items_len = util_array_len((const char **)items);

    for (i = 0; i < items_len; i++) {
        key_val = util_string_split(items[i], '=');
        if (key_val == NULL) {
            continue;
        }
        ret = parse_mounts_item(key_val[0], key_val[1], mount_str, m, errmsg);
        if (ret != 0) {
            goto out;
        }
        util_free_array(key_val);
        key_val = NULL;
    }

    if (m->type == NULL) {
#ifdef ENABLE_OCI_IMAGE
        m->type = util_strdup_s(DEFAULT_MOUNT_TYPE);
#else
        CACHE_ERRMSG(errmsg, "Invalid mount specification '%s'.Missing type", mount_str);
        ret = EINVALIDARGS;
        goto out;
#endif
    }

    if (check_mount_spec(mount_str, m, errmsg) != 0) {
        ret = EINVALIDARGS;
        goto out;
    }

    *spec = m;
    m = NULL;

out:
    if (ret != 0 && strlen(errmsg) != 0 && errmsg_out != NULL) {
        *errmsg_out = util_strdup_s(errmsg);
    }
    free_mount_spec(m);
    util_free_array(key_val);
    util_free_array(items);
    return ret;
}

bool util_valid_mount_spec(const char *mount_str, char **errmsg)
{
    int ret = 0;
    mount_spec *m = NULL;

    // if parse success, it's valid
    ret = util_parse_mount_spec((char *)mount_str, &m, errmsg);
    if (ret != 0) {
        goto out;
    }

out:
    free_mount_spec(m);

    return ret != 0 ? false : true;
}
